searchNodes=[{"ref":"Versioning.html","title":"Versioning","module":"Versioning","type":"module","doc":"Versionings allow data to be changed to different versions of itself. A the heart of our versioning is the Versioning struct. A Versioning struct contains the following fields: :current - The current version that our data represents. :target - The version that we want our data to be changed into. :type - The type of data we are working with. If we are working with structs, this will typically be the struct name in string format, eg: &quot;Post&quot; :data - The underlying data that we want to change. For structs, like our Post, be aware that we typically have our data as a bare map since it is easier to transform. :changes - A list of change modules that have been applied against the versioning. The first change module would be the most recent module run. :changed - A boolean representing if change modules have been applied. :assigns - A map of arbitrary data we can use to store additonal information in. Example Versioning.new(%Post{}, &quot;2.0.0&quot;, &quot;1.0.0&quot;) With the above, we have created a versioning of a Post struct. This represents us wanting to transform our post from a version &quot;2.0.0&quot; to an older &quot;1.0.0&quot; version. Schemas The versioning struct is used in combination with a Versioning.Schema, which allows us to map out the changes that should occur through versions. Please see the Versioning.Schema documentation for more details."},{"ref":"Versioning.html#assign/3","title":"Versioning.assign/3","module":"Versioning","type":"function","doc":"Assigns a value to a key in the versioning. The “assigns” storage is meant to be used to store values in the versioning so that change modules in your schema can access them. The assigns storage is a map. Examples iex&gt; versioning.assigns[:hello] nil iex&gt; versioning = Versioning.assign(versioning, :hello, :world) iex&gt; versioning.assigns[:hello] :world"},{"ref":"Versioning.html#fetch_data/2","title":"Versioning.fetch_data/2","module":"Versioning","type":"function","doc":"Fetches the value for a specific key in the data of versioning. If data contains the given key with value value, then {:ok, value} is returned. If data doesn&#39;t contain key, :error is returned. Examples iex&gt; Versioning.fetch_data(versioning, &quot;foo&quot;) {:ok, &quot;bar&quot;} iex&gt; Versioning.fetch_data(versioning, &quot;bar&quot;) :error"},{"ref":"Versioning.html#get_data/3","title":"Versioning.get_data/3","module":"Versioning","type":"function","doc":"Gets the value for a specific key in the data of versioning. If key is present in data with value value, then value is returned. Otherwise, default is returned (which is nil unless specified otherwise). Examples iex&gt; Versioning.get_data(versioning, &quot;foo&quot;) &quot;bar&quot; iex&gt; Versioning.get_data(versioning, &quot;bar&quot;) nil iex&gt; Versioning.get_data(versioning, &quot;bar&quot;, &quot;baz&quot;) &quot;baz&quot;"},{"ref":"Versioning.html#new/4","title":"Versioning.new/4","module":"Versioning","type":"function","doc":"Creates a new versioning using the data provided. If a struct is the data, and no type is provided, the struct module is set as the versioning :type (as described in put_type/2), and the struct is turned into a string-key map that is used for the :data. Examples # These are equivalent Versioning.new(%{&quot;foo&quot; =&gt; &quot;bar&quot;}, &quot;2.0.0&quot;, &quot;1.0.0&quot;, SomeData) Versioning.new(%{foo: &quot;bar&quot;}, &quot;2.0.0&quot;, &quot;1.0.0&quot;, &quot;SomeData&quot;) Versioning.new(%SomeData{foo: &quot;bar&quot;}, &quot;2.0.0&quot;, &quot;1.0.0&quot;)"},{"ref":"Versioning.html#pop_data/3","title":"Versioning.pop_data/3","module":"Versioning","type":"function","doc":"Returns and removes the value associated with key within the data of versioning. If key is present in data with value value, {value, new_versioning} is returned where new_versioning is the result of removing key from data. If key is not present in data, {default, new_versioning} is returned. Examples iex&gt; Versioning.pop_data(versioning, &quot;foo&quot;) {&quot;bar&quot;, versioning} iex&gt; Versioning.pop_data(versioning, &quot;foo&quot;) {nil, versioning}"},{"ref":"Versioning.html#put_current/2","title":"Versioning.put_current/2","module":"Versioning","type":"function","doc":"Puts the current version that the data represents. The version should be represented somewhere within your Versioning.Schema. This will become the &quot;starting&quot; point from which change modules will be run. Examples Versioning.put_current(versioning, &quot;0.1.0&quot;)"},{"ref":"Versioning.html#put_data/2","title":"Versioning.put_data/2","module":"Versioning","type":"function","doc":"Puts the full data in the versioning. The data represents the base of what will be modified when a versioning is run through a schema. Data must be a map. If a struct is provided, the struct will be turned into a basic map - though its type information will not be inferred. The keys of data will always be strings. If passed an Examples iex&gt; versioning = Versioning.put_data(versioning, %{&quot;foo&quot; =&gt; &quot;bar&quot;}) iex&gt; versioning.data %{&quot;foo&quot; =&gt; &quot;bar&quot;}"},{"ref":"Versioning.html#put_data/3","title":"Versioning.put_data/3","module":"Versioning","type":"function","doc":"Puts the given value under key within the data of versioning. Examples iex&gt; Versioning.put_data(versioning, &quot;foo&quot;, &quot;bar&quot;) iex&gt; versioning.data[&quot;foo&quot;] &quot;bar&quot;"},{"ref":"Versioning.html#put_target/2","title":"Versioning.put_target/2","module":"Versioning","type":"function","doc":"Puts the target version that the data will be transformed to. The version should be represented somewhere within your Versioning.Schema. Once the change modules in the target version are run, no more changes will be made. Examples Versioning.put_target(versioning, &quot;0.1.0&quot;)"},{"ref":"Versioning.html#put_type/2","title":"Versioning.put_type/2","module":"Versioning","type":"function","doc":"Puts the type of the versioning data. Typically, if working with data that is associated with a struct, this will be the struct trailing module name in binary format. For example, MyApp.Foo will be represented as &quot;Foo&quot;. When running a versioning through a schema, only the changes that match the type set on the versioning will be run. Examples # These are equivalent Versioning.put_type(versioning, &quot;Post&quot;) Versioning.put_type(versioning, MyApp.Post)"},{"ref":"Versioning.html#update_data/3","title":"Versioning.update_data/3","module":"Versioning","type":"function","doc":"Updates the key within the data of versioning using the given function. If the data does not contain key - nothing occurs. If it does, the fun is invoked with argument value and its result is used as the new value of key. Examples iex&gt; Versioning.update_data(versioning, &quot;foo&quot;, fn _val -&gt; &quot;bar&quot; end) iex&gt; versioning.data[&quot;foo&quot;] &quot;bar&quot;"},{"ref":"Versioning.html#t:assigns/0","title":"Versioning.assigns/0","module":"Versioning","type":"type","doc":""},{"ref":"Versioning.html#t:data/0","title":"Versioning.data/0","module":"Versioning","type":"type","doc":""},{"ref":"Versioning.html#t:t/0","title":"Versioning.t/0","module":"Versioning","type":"type","doc":""},{"ref":"Versioning.html#t:type/0","title":"Versioning.type/0","module":"Versioning","type":"type","doc":""},{"ref":"Versioning.html#t:version/0","title":"Versioning.version/0","module":"Versioning","type":"type","doc":""},{"ref":"Versioning.Change.html","title":"Versioning.Change","module":"Versioning.Change","type":"behaviour","doc":"Defines a versioning change. A versioning change is used to make small changes to data of a certain type. They are used within a Versioning.Schema. Changes should attempt to be as focused as possible to ensure complexity is kept to a minimum. Example defmodule MyApp.Cheanges.PostStatusChange do use Versioning.Change @desc &quot;The &#39;active&#39; attribute has been changed in favour of the &#39;status&#39; attribute&quot; @impl Versioning.Change def down(versioning, _opts) do case Versioning.pop_data(versioning, &quot;status&quot;) do {:active, versioning} -&gt; Versioning.put_data(versioning, &quot;active&quot;, true) {_, versioning} -&gt; Versioning.put_data(versioning, &quot;active&quot;, false) end end @impl Versioning.Change def up(versioning, _opts) do case Versioning.pop_data(versioning, &quot;active&quot;) do {true, versioning} -&gt; Versioning.put_data(versioning, &quot;status&quot;, &quot;active&quot;) {false, versioning} -&gt; Versioning.put_data(versioning, &quot;status&quot;, &quot;hidden&quot;) {_, versioning} -&gt; versioning end end end The above change module represents us modifying our Post data to support a new attribute - status - which replaces the previous active attribute. When changing data &quot;down&quot;, we must remove the status attribte, and replace it with a value that represents the previous active attribute. When changing data &quot;up&quot;, we must remove the active attribute and replace it with a value that represents the new status attribute. Descriptions Change modules can optionally include a @desc module attribute. This will be used to describe the changes made in the change module when constructing changelogs. Please see the Versioning.Changelog documentation for more information on changelogs."},{"ref":"Versioning.Change.html#c:down/2","title":"Versioning.Change.down/2","module":"Versioning.Change","type":"callback","doc":"Accepts a Versioning struct and applies changes downward. Examples MyApp.Change.down(versioning)"},{"ref":"Versioning.Change.html#c:up/2","title":"Versioning.Change.up/2","module":"Versioning.Change","type":"callback","doc":"Accepts a Versioning struct, and applies changes upward. Examples MyApp.Change.up(versioning)"},{"ref":"Versioning.ChangelogError.html","title":"Versioning.ChangelogError","module":"Versioning.ChangelogError","type":"exception","doc":""},{"ref":"Versioning.CompileError.html","title":"Versioning.CompileError","module":"Versioning.CompileError","type":"exception","doc":""},{"ref":"Versioning.MissingSchemaError.html","title":"Versioning.MissingSchemaError","module":"Versioning.MissingSchemaError","type":"exception","doc":""},{"ref":"Versioning.MissingVersionError.html","title":"Versioning.MissingVersionError","module":"Versioning.MissingVersionError","type":"exception","doc":""},{"ref":"Versioning.Schema.html","title":"Versioning.Schema","module":"Versioning.Schema","type":"module","doc":"Defines a versioning schema. A versioning schema is used to change data through a series of steps from a &quot;current&quot; version to a &quot;target&quot; version. This is useful in maintaining backwards compatability with older versions of API&#39;s without enormous complication. Example defmodule MyApp.Versioning do use Versioning.Schema, adapter: Versioning.Adapter.Semantic version(&quot;2.0.0&quot;, do: []) version &quot;1.1.0&quot; do type &quot;User&quot; do change(MyApp.Changes.SomeUserChange) end end version &quot;1.0.1&quot; do type &quot;Post&quot; do change(MyApp.Changes.SomePostChange)) end type &quot;All!&quot; do change(MyApp.Changes.SomeAllChange) end end version(&quot;1.0.0&quot;, do: []) end When creating a schema, an adapter must be specified. The adapter determines how versions are parsed and compared. For more information on adapters, please see Versioning.Adapter. In the example above, we have 4 versions. Our current version is represented by the top version - &quot;2.0.0&quot;. Our oldest version is at the bottom - &quot;1.0.0&quot;. We define a version with the version/2 macro. Within a version, we specify types that have been manipulated. We define a type with the type/2 macro. Within a type, we specify changes that have occured. We define a change with the change/2 macro. Running Schemas Lets say we have a %Post{} struct that we would like to run through our schema. post = %Post{status: :enabled} Versioning.new(post, &quot;2.0.0&quot;, &quot;1.0.0&quot;) We have created a new versioning of our post struct. The versioning sets the data, current version, target version, as well as type. We can now run our versioning through our schema. {:ok, versioning} = MyApp.Versioning.run(versioning) With the above, our versioning struct will first be run through our MyApp.Changes.SomePostChange change module as the type matches our versioning type. It will then be run through our MyApp.Changes.SomeAllChange as it also matches on the &quot;All! type (more detail available at the change/2 macro). With the above, we are transforming our data &quot;down&quot; through our schema. But we can also transform it &quot;up&quot;. post = %{&quot;status&quot; =&gt; &quot;some_status&quot;} Versioning.new(post, &quot;1.0.0&quot;, &quot;2.0.0&quot;, &quot;Post&quot;) If we were to run our new versioning through the schema, the same change modules would be run, but in reverse order. Change Modules At the heart of versioning schemas are change modules. You can find more information about creating change modules at the Versioning.Change documentation. Schema attributes Supported attributes for configuring the defined schema. They must be set after the use Versioning.Schema call. These attributes are: @latest - configures the schema latest version. By default, this will be the version at the top of your schema. But if you do not wish to have this behaviour, you can set it here. Reflection Any schema module will generate the __schema__ function that can be used for runtime introspection of the schema: __schema__(:down) - Returns the data structure representing a downward versioning. __schema__(:up) - Returns the data structure representing an upward versioning. __schema__(:adapter) - Returns the versioning adapter used by the schema. __schema__(:latest, :string) - Returns the latest version in string format. __schema__(:latest, :parsed) - Returns the latest version in parsed format."},{"ref":"Versioning.Schema.html#change/2","title":"Versioning.Schema.change/2","module":"Versioning.Schema","type":"macro","doc":"Defines a change within a type. A change must be represented by a module that implements the Versioning.Change behaviour. You can also set options that will be passed along to the change module. Changes are run in the order they are placed, based on the direction of the version change. For instance, if a schema was being run &quot;down&quot; for the example below, MyChangeModule would be run first, followed by MyOtherChangeModule. This would be reversed if running &quot;up&quot; a schema. Example version &quot;1.0.1&quot; do type &quot;Foo&quot; do change(MyChangeModule) change(MyOtherChangeModule, [foo: :bar]) end end"},{"ref":"Versioning.Schema.html#type/2","title":"Versioning.Schema.type/2","module":"Versioning.Schema","type":"macro","doc":"Defines a type within a version. A type can only be represented once within a version, and must be a string. Any issue with this will raise a Versioning.CompileError during compilation. Typically, it should be represented in &quot;CamelCase&quot; format. Any changes within a type that matches the type on a Versioning struct will be run. There is also the special case &quot;All!&quot; type, which lets you define changes that will be run against all versionings - regardless of type. Example version &quot;1.0.1&quot; do type &quot;All!&quot; do end type &quot;Foo&quot; do end end"},{"ref":"Versioning.Schema.html#version/2","title":"Versioning.Schema.version/2","module":"Versioning.Schema","type":"macro","doc":"Defines a version in the schema. A version must be in string format, and must adhere to requirements of the Elixir Version module. This means SemVer 2.0. A version can only be represented once within a schema. The most recent version should be at the top of your schema, and the oldest at the bottom. Any issue with the above will raise a Versioning.CompileError during schema compilation. Example version &quot;1.0.1&quot; do end version(&quot;1.0.0&quot;, do: [])"},{"ref":"Versioning.Schema.html#t:change/0","title":"Versioning.Schema.change/0","module":"Versioning.Schema","type":"type","doc":""},{"ref":"Versioning.Schema.html#t:direction/0","title":"Versioning.Schema.direction/0","module":"Versioning.Schema","type":"type","doc":""},{"ref":"Versioning.Schema.html#t:result/0","title":"Versioning.Schema.result/0","module":"Versioning.Schema","type":"type","doc":""},{"ref":"Versioning.Schema.html#t:schema/0","title":"Versioning.Schema.schema/0","module":"Versioning.Schema","type":"type","doc":""},{"ref":"Versioning.Schema.html#t:t/0","title":"Versioning.Schema.t/0","module":"Versioning.Schema","type":"type","doc":""},{"ref":"Versioning.Schema.html#t:type/0","title":"Versioning.Schema.type/0","module":"Versioning.Schema","type":"type","doc":""},{"ref":"Versioning.Schema.html#t:version/0","title":"Versioning.Schema.version/0","module":"Versioning.Schema","type":"type","doc":""},{"ref":"VersioningError.html","title":"VersioningError","module":"VersioningError","type":"exception","doc":""},{"ref":"Versioning.Adapter.html","title":"Versioning.Adapter","module":"Versioning.Adapter","type":"behaviour","doc":"Defines a versioning adapter. A versioning adapter is used to parse and compare versions. This allows versions to be defined in a variety of ways - from semantic, to date based."},{"ref":"Versioning.Adapter.html#c:compare/2","title":"Versioning.Adapter.compare/2","module":"Versioning.Adapter","type":"callback","doc":"Callback invoked to compare versions. Returns :gt if the first verison is greater than the second, and :lt for vice-versa. If the two versions are equal, :eq is returned."},{"ref":"Versioning.Adapter.html#c:parse/1","title":"Versioning.Adapter.parse/1","module":"Versioning.Adapter","type":"callback","doc":"Callback invoked to parse a binary version. Returns {:ok, term} on success, where term will be the adapters representation of a version. Returns :error if the version cannot be parsed."},{"ref":"Versioning.Adapter.html#t:t/0","title":"Versioning.Adapter.t/0","module":"Versioning.Adapter","type":"type","doc":""},{"ref":"Versioning.Adapter.Date.html","title":"Versioning.Adapter.Date","module":"Versioning.Adapter.Date","type":"module","doc":"A versioning adapter for date-based versions. Under the hood, this adapter uses the Date module. For details on the rules that are used for parsing and comparison, please see the Date module. Example defmodule MyApp.Versioning do use Versioning.Schema, adapter: Versioning.Adapter.Date version &quot;2019-01-01&quot; do type &quot;Post&quot; do change(MyApp.Change) end end end"},{"ref":"Versioning.Adapter.Date.html#compare/2","title":"Versioning.Adapter.Date.compare/2","module":"Versioning.Adapter.Date","type":"function","doc":"Compares date based versions using ISO8601 formatting. Returns :gt if the first verison is greater than the second, and :lt for vice-versa. If the two versions are equal, :eq is returned. Returns :error if the version cannot be parsed. Example iex&gt; Versioning.Adapters.Date.compare(&quot;2019-01-01&quot;, &quot;2018-12-31&quot;) :gt iex&gt; Versioning.Adapters.Date.compare(&quot;2018-12-31&quot;, &quot;2019-01-01&quot;) :lt iex&gt; Versioning.Adapters.Date.compare(&quot;2019-01-01&quot;, &quot;2019-01-01&quot;) :eq iex&gt; Versioning.Adapters.Date.compare(&quot;foo&quot;, &quot;bar&quot;) :error"},{"ref":"Versioning.Adapter.Date.html#parse/1","title":"Versioning.Adapter.Date.parse/1","module":"Versioning.Adapter.Date","type":"function","doc":"Parses date based versions using ISO8601 formatting. Example iex&gt; Versioning.Adapters.Date.parse(&quot;2019-01-01&quot;) {:ok, ~D[2019-01-01]} iex&gt; Versioning.Adapters.Date.parse(&quot;foo&quot;) :error"},{"ref":"Versioning.Adapter.Semantic.html","title":"Versioning.Adapter.Semantic","module":"Versioning.Adapter.Semantic","type":"module","doc":"A versioning adapter for semantic-based versions. Under the hood, this adapter uses the Version module. For details on the rules that are used for parsing and comparison, please see the Version module. Example defmodule MyApp.Versioning do use Versioning.Schema, adapter: Versioning.Adapter.Semantic version &quot;1.0.0&quot; do type &quot;Post&quot; do change(MyApp.Change) end end end"},{"ref":"Versioning.Adapter.Semantic.html#compare/2","title":"Versioning.Adapter.Semantic.compare/2","module":"Versioning.Adapter.Semantic","type":"function","doc":"Compares semantic based versions. Returns :gt if the first verison is greater than the second, and :lt for vice-versa. If the two versions are equal, :eq is returned. Returns :error if the version cannot be parsed. Example iex&gt; Versioning.Adapter.Semantic.compare(&quot;1.0.1&quot;, &quot;1.0.0&quot;) :gt iex&gt; Versioning.Adapter.Semantic.compare(&quot;1.0.0&quot;, &quot;1.0.1) :lt iex&gt; Versioning.Adapter.Semantic.compare(&quot;1.0.1&quot;, &quot;1.0.1&quot;) :eq iex&gt; Versioning.Adapter.Semantic.compare(&quot;foo&quot;, &quot;bar&quot;) :error"},{"ref":"Versioning.Adapter.Semantic.html#parse/1","title":"Versioning.Adapter.Semantic.parse/1","module":"Versioning.Adapter.Semantic","type":"function","doc":"Parses semantic based versions. Example iex&gt; Versioning.Adapter.Semantic.parse(&quot;1.0.0&quot;) {:ok, #Version&lt;1.0.0&gt;} iex&gt; Versioning.Adapter.Semantic.parse(&quot;foo&quot;) :error"},{"ref":"Versioning.Changelog.html","title":"Versioning.Changelog","module":"Versioning.Changelog","type":"module","doc":"Creates changelogs for schemas. The changelog is composed of a list of maps that describe the history of the schema. For example: [ %{ version: &quot;1.1.0&quot;, changes: [ %{type: Foo, descriptions: [&quot;Changed this.&quot;, &quot;Changed that.&quot;]} ] }, %{ version: &quot;1.0.0&quot;, changes: [ %{type: Foo, descriptions: [&quot;Changed this.&quot;, &quot;Changed that.&quot;]} ] } ] The descriptions associated with each change can be attributed via the @desc module attribute on a change module. Please see Versioning.Change for more details. Formatters can be used to turn the raw changelog into different formats. Please see the Versioning.Changelog.Formatter behaviour for more details. The Versioning.Changelog.Markdown formatter is included with this package."},{"ref":"Versioning.Changelog.html#build/2","title":"Versioning.Changelog.build/2","module":"Versioning.Changelog","type":"function","doc":"Builds a changelog of the schema. Options :version - A specific version within the changelog. :type - A specific type within the specified version. :formatter - A module that adheres to the Versioning.Changelog.Formatter behaviour. Examples Versioning.Changelog.build(MySchema, formatter: Versioning.Changelog.Markdown)"},{"ref":"Versioning.Changelog.html#t:change/0","title":"Versioning.Changelog.change/0","module":"Versioning.Changelog","type":"type","doc":""},{"ref":"Versioning.Changelog.html#t:t/0","title":"Versioning.Changelog.t/0","module":"Versioning.Changelog","type":"type","doc":""},{"ref":"Versioning.Changelog.html#t:version/0","title":"Versioning.Changelog.version/0","module":"Versioning.Changelog","type":"type","doc":""},{"ref":"Versioning.Changelog.Formatter.html","title":"Versioning.Changelog.Formatter","module":"Versioning.Changelog.Formatter","type":"behaviour","doc":"Defines a versioning changelog formatter. A changelog formatter is used to create custom outputs from raw changelog data. Included with this package is the Versioning.Changelog.Markdown ormatter. This accepts the standard changelog data structure, and converts it to a simple markdown format. Example defmodule MyApp.SomeFormatter do use Versioning.Changelog.Formatter @impl Versioning.Changelog.Formatter def format(changelog) do # Do custom formatting end end Please see the Versioning.Changelog.Markdown for an example of its use."},{"ref":"Versioning.Changelog.Formatter.html#c:format/1","title":"Versioning.Changelog.Formatter.format/1","module":"Versioning.Changelog.Formatter","type":"callback","doc":"Formats a changelog. Accepts a list of changelog versions, a single version, or a single change, and returns a custom formatted version."},{"ref":"Versioning.Changelog.Markdown.html","title":"Versioning.Changelog.Markdown","module":"Versioning.Changelog.Markdown","type":"module","doc":"Formats a changelog into a markdown version."},{"ref":"Versioning.Controller.html","title":"Versioning.Controller","module":"Versioning.Controller","type":"module","doc":"A set of functions used with Phoenix controllers. Typically, this module should be imported into your controller modules. In a normal phoenix application, this can usually be done with the following: defmodule YourAppWeb do # ... def controller do quote do use Phoenix.Controller, namespace: MyAppWeb # ... import Versioning.Controller # ... end end end Please see the documentation at Phoenix.Controller for details on how to set up a typical controller. This module is mainly used to convert raw params from the version that the request data represents, to the latest version that your &quot;context&quot; application expects. Example Below is an example of how to use versioning in a typical controller: defmodule MyController do use MyAppWeb, :controller plug Versioning.Plug, schema: MyVersioningSchema def update(conn, %{&quot;id&quot; =&gt; id, &quot;user&quot; =&gt; params}) do with {:ok, params} &lt;- params_version(conn, params, &quot;User&quot;), {:ok, user} &lt;- Blog.fetch_user(id), {:ok, user} &lt;- Blog.update_user(user) do render(conn, &quot;show.json&quot;, user: user) end end end The params_version/3 function accepts a conn, a set of params representing whatever version of the data the user uses, as well as the type of the data. It will run the params through your schema, returning the results."},{"ref":"Versioning.Controller.html#apply_version/3","title":"Versioning.Controller.apply_version/3","module":"Versioning.Controller","type":"function","doc":"Applies a version using the header or fallback. The schema must already be stored on the conn to use this function. The fallback is used if the header is not present. Its value can be :latest, representing the latest version on the schema, or a {module, function}. This module and function will be called with the conn."},{"ref":"Versioning.Controller.html#fetch_schema/1","title":"Versioning.Controller.fetch_schema/1","module":"Versioning.Controller","type":"function","doc":"Fetches the current schema. Returns {:ok, schema} on success, or :error if no schema exists. Examples iex&gt; conn = Versioning.Controller.put_schema(conn, MySchema) iex&gt; Versioning.Controller.fetch_schema(conn) {:ok, MySchema}"},{"ref":"Versioning.Controller.html#fetch_schema!/1","title":"Versioning.Controller.fetch_schema!/1","module":"Versioning.Controller","type":"function","doc":"Fetches the current schema or errors if empty. Returns schema or raises a Versioning.MissingSchemaError. Examples iex&gt; conn = Versioning.Controller.put_schema(conn, MySchema) iex&gt; Versioning.Controller.fetch_schema!(conn) MySchema"},{"ref":"Versioning.Controller.html#fetch_version/1","title":"Versioning.Controller.fetch_version/1","module":"Versioning.Controller","type":"function","doc":"Fetches the current request version. Returns {:ok, version} on success, or :error if no version exists. Examples iex&gt; conn = Versioning.Controller.put_version(conn, &quot;1.0.0&quot;) iex&gt; Versioning.Controller.fetch_version(conn) {:ok, &quot;1.0.0&quot;}"},{"ref":"Versioning.Controller.html#fetch_version!/1","title":"Versioning.Controller.fetch_version!/1","module":"Versioning.Controller","type":"function","doc":"Fetches the current request version or errors if empty. Returns version or raises a Versioning.MissingVersionError. Examples iex&gt; conn = Versioning.Controller.put_schema(conn, &quot;1.0.0&quot;) iex&gt; Versioning.Controller.fetch_version!(conn) &quot;1.0.0&quot;"},{"ref":"Versioning.Controller.html#params_version/3","title":"Versioning.Controller.params_version/3","module":"Versioning.Controller","type":"function","doc":"Performs versioning on the params using the given type. The schema and request version must already be stored on the conn to use this function. Returns {:ok, params} with the new versioned params, or {:error, :bad_version} if the schema does not contain the version requested."},{"ref":"Versioning.Controller.html#put_schema/2","title":"Versioning.Controller.put_schema/2","module":"Versioning.Controller","type":"function","doc":"Stores the schema for versioning. Examples Versioning.Controller.put_schema(conn, MySchema)"},{"ref":"Versioning.Controller.html#put_version/2","title":"Versioning.Controller.put_version/2","module":"Versioning.Controller","type":"function","doc":"Stores the request version. Examples Versioning.Controller.put_version(conn, &quot;1.0.0&quot;)"},{"ref":"Versioning.Plug.html","title":"Versioning.Plug","module":"Versioning.Plug","type":"module","doc":"A Plug to assist with versioning an API. It requires one option: :schema- the versioning schema to use when changing data. We can easily add our plug to a pipeline with the following: plug Versioning.Plug, schema: MySchema The above will store the schema within the conn struct, as well as attempt to apply a version to the request. This information can then be used when changing data to and from the latest version. Options :header - the header to read when getting the version requested. This defaults to &quot;x-api-version&quot;. :fallback - the fallback to occur if no version can be found in the header. This defaults to :latest - which is the latest version in our schema. Alternatively, you can specify a {module, function}. This enables more dynamic behaviour - such as fetching a version from a current user or token. The module and function must have an arity of 1 - the conn struct will be passed to it. Examples This plug can be mounted in a Plug.Builder pipeline as follows: defmodule MyPlug do use Plug.Builder plug Versioning.Plug, schema: MySchema, header: &quot;myapi-version&quot;, fallback: {MyFallback, :call} plug :not_found def not_found(conn, _) do send_resp(conn, 404, &quot;not found&quot;) end end"},{"ref":"Versioning.View.html","title":"Versioning.View","module":"Versioning.View","type":"module","doc":"A set of functions used with Phoenix views. Typically, this module should be imported into your view modules. In a normal phoenix application, this can usually be done with the following: defmodule YourAppWeb do # ... def view do quote do use Phoenix.View, root: &quot;lib/your_app_web/templates&quot;, namespace: &quot;web&quot; # ... import Versioning.View # ... end end end Please see the documentation at Phoenix.View for details on how to set up a typical view. In places that you would use Phoenix.View.render_one/4, this module provides render_version/6. In places that you would use Phoenix.View.render_many/4, this module provides render_versions/6. In order to use these functions, you must already have applied the schema and requested version to the conn. This is typically done with Versioning.Plug or through the helpers available in Versioning.Controller. Example Below is an example of how to use versioning in a typical view: defmodule YourApp.UserView do use YourApp.View def render(&quot;index.json&quot;, %{conn: conn, users: users}) do %{ &quot;users&quot; =&gt; render_versions(conn, users, &quot;User&quot;, UserView, &quot;user.json&quot;), } end def render(&quot;show.json&quot;, %{conn: conn, user: user}) do %{ &quot;user&quot; =&gt; render_version(conn, users, &quot;User&quot;, UserView, &quot;user.json&quot;), } end def render(&quot;user.json&quot;, %{user: user}) do %{&quot;name&quot; =&gt; user.name, &quot;address&quot; =&gt; user.address} end end A typical call, such as: render_many(users, UserView, &quot;user.json&quot;) Is replaced by the following: render_versions(conn, users, &quot;User&quot;, UserView, &quot;user.json&quot;) In order to render versions of our data, we must pass the conn struct, our data to be versioned, the type the data represents in our schema, the view module to use, the template to use, as well as an additional assigns. The contents of the &quot;user.json&quot; template represent the latest version of your data. They will be run through your versioning schema to the version requested by the user. The output returned by your schema is what will be finally rendered."},{"ref":"Versioning.View.html#render_version/6","title":"Versioning.View.render_version/6","module":"Versioning.View","type":"function","doc":"Renders a single versioned item if not nil. render_version(conn, user, &quot;User&quot;, UserView, &quot;show.json&quot;) This require Under the hood, this will render the item using Phoenix.View.render/3 - so the latest version of the data should be represented in your view using typical view standards. After the data has been rendered, it will be passed to your schema and versioned to requested target version."},{"ref":"Versioning.View.html#render_versions/6","title":"Versioning.View.render_versions/6","module":"Versioning.View","type":"function","doc":"Renders a versioned collection. A collection is any enumerable of structs. This function returns the rendered versioned collection in a list: render_versions(conn, users, &quot;User&quot;, UserView, &quot;show.json&quot;) Under the hood, this will render each item using Phoenix.View.render/3 - so the latest version of the data should be represented in your view using typical view standards. After the data has been rendered, it will be passed to your schema and versioned to the version that has been requested."},{"ref":"readme.html","title":"Versioning","module":"readme","type":"extras","doc":"Versioning Versioning provides a way for API&#39;s to remain backward compatible without the headache. This is done through use of a &quot;versioning schema&quot; that translates data through a series of steps from its current version to the target version. This technique is well described in the article APIs as infrastructure: future-proofing Stripe with versioning. The basic rule is each API version in the schema must only ever concern itself with creating a set of change modules associated with the version below/above it. This contract ensures that we can continue to translate data to legacy versions without enormous effort."},{"ref":"readme.html#installation","title":"Versioning - Installation","module":"readme","type":"extras","doc":"The package can be installed by adding versioning to your list of dependencies in mix.exs: def deps do [ {:versioning, &quot;~&gt; 0.3&quot;} ] end"},{"ref":"readme.html#documentation","title":"Versioning - Documentation","module":"readme","type":"extras","doc":"See HexDocs for additional documentation."},{"ref":"readme.html#example","title":"Versioning - Example","module":"readme","type":"extras","doc":"For a more in-depth example, please check out the Getting Started page. We build a schema to describe updates to our API through versions, types, and changes. defmodule MyAPI.Versioning do use Versioning.Schema version(&quot;1.2.0&quot;, do: []) version &quot;1.1.0&quot; do type &quot;Post&quot; do change(MyAPI.Changes.PostStatusChange) end end version(&quot;1.0.0&quot;, do: []) end We build a change module to perform data modifications. defmodule MyAPI.Changes.PostStatusChange do use Versioning.Change @desc &quot;&quot;&quot; The boolean field &quot;active&quot; was removed in favour of the enum &quot;status&quot;. &quot;&quot;&quot; def down(versioning, _opts) do case Versioning.pop_data(versioning, &quot;status&quot;) do {&quot;active&quot;, versioning} -&gt; Versioning.put_data(versioning, &quot;active&quot;, true) {_, versioning} -&gt; Versioning.put_data(versioning, &quot;active&quot;, false) end end def up(versioning, _opts) do case Versioning.pop_data(versioning, &quot;active&quot;) do {true, versioning} -&gt; Versioning.put_data(versioning, &quot;status&quot;, &quot;active&quot;) {false, versioning} -&gt; Versioning.put_data(versioning, &quot;status&quot;, &quot;hidden&quot;) {_, versioning} -&gt; versioning end end end We create versionings to run against our schema. Here, we want to change our post data from version 1.2.0 to 1.0.0. We can then run our versioning against the schema, which will return a modified versioning with our change modules run. versioning = Versioning.new(%Post{}, &quot;1.2.0&quot;, &quot;1.0.0&quot;) MyAPI.Versioning.run(versioning)"},{"ref":"readme.html#phoenix-usage","title":"Versioning - Phoenix Usage","module":"readme","type":"extras","doc":"Versioning provides extensive support for usage with Phoenix. Checkout the Phoenix Usage page for more details."},{"ref":"getting-started.html","title":"Getting Started","module":"getting-started","type":"extras","doc":"Getting Started Lets say we have a Post struct that contains the boolean field :active. As time goes by, we recognize that there may be more kinds of statuses that our Post&#39;s may have. To keep up with the times, we replace our :active field with the enum field :status. One of the values could be &quot;active&quot; - among many others. We arent ready to release this feature into the the wild yet though. So, while we change the internal structure of our Post data, we must ensure we dont break the API contract we made with our users. Versioning to the rescue..."},{"ref":"getting-started.html#versioning-struct","title":"Getting Started - Versioning Struct","module":"getting-started","type":"extras","doc":"A the heart of our versioning is the Versioning struct. A Versioning struct contains the following fields: :current - The current version that our data represents. :target - The version that we want our data to be changed into. :type - The type of data we are working with. If we are working with structs, this will typically be the struct name, eg: Post :data - The underlying data that we want to change. For structs, like our Post, be aware that we typically have our data as a bare map since it is easier to transform. :changed - A boolean representing whether a change operation has occured. :assigns - A map of arbitrary data we can use to store additonal information in. Let&#39;s see a couple different ways we can create a versioning a Post. # The type is automatically inferred from the struct module. Versioning.new(%Post{}, &quot;2.0.0&quot;, &quot;1.0.0&quot;) # We can also explicely set the type. Versioning.new(%{}, &quot;2.0.0&quot;, &quot;1.0.0&quot;, Post) # We can also build up our versioning using helpers. |&gt; Versioning.new() |&gt; Versioning.put_data(post) |&gt; Versioning.put_type(Post) |&gt; Versioning.put_current(&quot;2.0.0&quot;) |&gt; Versioning.put_target(&quot;1.0.0&quot;) We now have a versioning of our Post."},{"ref":"getting-started.html#versioning-change","title":"Getting Started - Versioning Change","module":"getting-started","type":"extras","doc":"Lets create of first &quot;versioning change&quot;. This is a module that adheres to the Versioning.Change behaviour. From it, we must implement the callbacks up/2 and down/2. defmodule MyAPI.Changes.PostStatusChange do use Versioning.Change @desc &quot;&quot;&quot; The boolean field &quot;active&quot; was removed in favour of the enum &quot;status&quot;. &quot;&quot;&quot; def down(versioning, _opts) do case Versioning.pop_data(versioning, :status) do {:active, versioning} -&gt; Versioning.put_data(versioning, :active, true) {_, versioning} -&gt; Versioning.put_data(versioning, :active, false) end end def up(versioning, _opts) do case Versioning.pop_data(versioning, &quot;active&quot;) do {true, versioning} -&gt; Versioning.put_data(versioning, &quot;status&quot;, &quot;active&quot;) {false, versioning} -&gt; Versioning.put_data(versioning, &quot;status&quot;, &quot;hidden&quot;) {_, versioning} -&gt; versioning end end end Our down/2 function accepts a versioning, removes the new :status value, and translates it to the old :active requirements - returning a modified versioning. Our up/2 function accepts a versioning, removes the old &quot;active&quot; value, and translates it to our new &quot;status&quot; requirements - returning a modified versioning. We can also use the @desc module attribute to attach a description of the change. This will be used when generating a changelog."},{"ref":"getting-started.html#versioning-schema","title":"Getting Started - Versioning Schema","module":"getting-started","type":"extras","doc":"With our first change module in place, its time to tie it all together with our &quot;versioning schema&quot;. The schema provides a DSL to describe and route our versioning. defmodule MyAPI.Versioning do use Versioning.Schema version(&quot;2.0.0&quot;, do: []) version &quot;1.1.0&quot; do type &quot;Post&quot; do change(MyAPI.Changes.PostStatusChange) end end version(&quot;1.0.0&quot;, do: []) end The schema above shows we currently support 3 versions. Our top version &quot;2.0.0&quot; represents the current version. &quot;1.1.0&quot; is where our new article change is held. The schema DSL describes a flow, whereby the &quot;top&quot; version represents the most recent, and each subsequent version is one step older."},{"ref":"getting-started.html#running-our-versioning","title":"Getting Started - Running our Versioning","module":"getting-started","type":"extras","doc":"With our versioning in place, we can now translate our Post struct to the requirements of our users &quot;pinned&quot; API version. #For the sake of example, lets say the user is pinned at the older &quot;1.0.0&quot; version. version = get_api_version(user) post = get_post(id) versioning = Versioning.new(post, &quot;2.0.0&quot;, version) MyAPI.Versioning.run(versioning) #Versioning&lt;current: &quot;2.0.0&quot;, target: &quot;1.0.0&quot;, type: Post, changed: true&gt; Calling run/1 on our schema with a versioning struct will execute our schema downwards/upwards depending on the order of our current and target versions. It will &quot;walk&quot; through each version, running any changes held within it that match the :type on our versioning struct. A schema is typically run &quot;downwards&quot; when converting local data to external data. A schema is typically run &quot;updwards&quot; when converting external data to local data. Once a matching version is found, it will run the changes within, but will stop execution afterwards. We can then access the underlying data through our versioning.data."},{"ref":"getting-started.html#versioning-changelog","title":"Getting Started - Versioning Changelog","module":"getting-started","type":"extras","doc":"A changelog of our schema can also be generated. This changelog represents a list of maps in the format (shortend for brevity): [ %{ version: &quot;2.0.0&quot;, changes: [] }, %{ version: &quot;1.1.0&quot;, changes: [ %{ type: Post, descriptions: [ &quot;The boolean field `:active` was removed in favour of the enum `:status`.&quot; ] } ] }, %{ version: &quot;1.0.0&quot;, changes: [] } ] You can access the changelog while providing options such as a formatter. Included with Versioning is a basic markdown formatter. Versioning.Changelog.build(MyAPI.Versioning, formatter: Versioning.Changelog.Markdown)"},{"ref":"phoenix-usage.html","title":"Phoenix Usage","module":"phoenix-usage","type":"extras","doc":"Phoenix Usage Versioning provides extensive support for its usage with Phoenix. The best way to get started is by adding some helper functions to your controllers and views. defmodule YourAppWeb do # ... def controller do quote do use Phoenix.Controller, namespace: MyAppWeb # ... import Versioning.Controller # ... end end def view do quote do use Phoenix.View, root: &quot;lib/your_app_web/templates&quot;, namespace: &quot;web&quot; # ... import Versioning.View # ... end end end Followed by adding a plug to your endpoint. Typically, you&#39;ll want this plug to be added after you have access to the &quot;current user&quot;. defmodule YourAppWeb.Endpoint do use Phoenix.Endpoint, otp_app: :your_app # plug ... plug Versioning.Plug, schema: YourAppWeb.Versioning plug YourAppWeb.Router end Beyond the above, you&#39;ll of course need to setup your versioning schema and changes. But with the above, you should be good to go! Please consult the documentation available at Versioning.Controller for help on how to version params. And please see the documentation available at Versioning.View for help on how to version responses."}]